#version 300 es

uniform lowp sampler2D diffuseMap;
uniform lowp sampler2D normalMap;
uniform lowp sampler2D specularMap;
uniform lowp samplerCube envMap;
uniform lowp sampler2D _ShadowMap;
uniform lowp vec3 materialSpecular;
uniform lowp vec3 _LightColor;
uniform lowp float roughness;
uniform lowp float reflectivity;

in highp vec2 vTexCoord;
in highp vec3 vViewVec;
in highp vec3 vLightVec;
#ifdef ENVIRONMENT_MAP
in highp vec3 vEnvTexCoord;
#endif

out lowp vec4 fragColor;

// Shadow stuff
in highp vec4 vPosition;
in highp vec4 vProjCoord;
uniform highp float uLinearDepthConstant;
uniform highp vec3 _LightPositionWorld;
uniform highp int _LightCastsShadow;

highp float linstep( highp float low, highp float high, highp float v )
{
    return clamp( (v - low) / (high - low), 0.0, 1.0 );
}

highp float VSM( in highp float depth )
{
    highp vec2 moments = texture( _ShadowMap, vProjCoord.xy ).rg;
    
    highp float variance = moments.y - moments.x * moments.x;
    //variance = max( variance, 0.0005 );
    
    highp float delta = depth - moments.x;
    highp float pMax = linstep( 0.2, 1.0, variance / (variance + delta * delta) );
    
    return (depth <= moments.x) ? 1.0 : pMax;
}

void main()
{
    mediump vec3 albedoTex = texture( diffuseMap, vTexCoord ).rgb;
#ifdef SPECULAR_MAP
    mediump vec3 specularTex = texture( specularMap, vTexCoord ).rgb;
#else
    mediump vec3 specularTex = materialSpecular;
#endif
    
    // input vectors
    mediump vec3 N = normalize( texture( normalMap, vTexCoord ).rgb * 2.0 - 1.0 );
    mediump vec3 V = normalize( vViewVec );
    mediump vec3 L = normalize( vLightVec );
    mediump vec3 H = normalize( L + V );
    
#ifdef ENVIRONMENT_MAP
    mediump vec3 envTex = texture( envMap, vEnvTexCoord + N.xyz ).rgb;
    albedoTex = mix( albedoTex, envTex, reflectivity ) * specularTex; // Could use a reflection mask texture or channel.
#endif
    
    // dot products
    mediump float dotNL = clamp( dot( N, L ), 0.0, 1.0 );
    mediump float dotNH = clamp( dot( N, H ), 0.0, 1.0 );
    mediump float dotNV = clamp( dot( N, V ), 0.0, 1.0 );
    mediump float dotVH = clamp( dot( V, H ), 0.0, 1.0 );
    mediump float r_sq = roughness * roughness;
    
    // Geometric term.
    mediump float geoNumerator = 2.0 * dotNH;
    mediump float geoB = (geoNumerator * dotNV) / dotVH;
    mediump float geoC = (geoNumerator * dotNL) / dotVH;
    mediump float geo = min( 1.0, min( geoB, geoC ) );
    
    // Beckmann roughness.
    mediump float roughness_a = 1.0 / (4.0 * r_sq * pow( dotNH, 4.0 ) );
    mediump float roughness_b = dotNH * dotNH - 1.0;
    mediump float roughness_c = r_sq * dotNH * dotNH;
    mediump float roughness = roughness_a * exp( roughness_b / roughness_c );

    // Schlick Fresnel.
    mediump float ref_at_norm_incidence = 1.0;
    mediump float fresnel = pow( 1.0 - dotVH, 5.0 );
    fresnel *= (1.0 - ref_at_norm_incidence);
    fresnel += ref_at_norm_incidence;
    
    // Specular BRDF.
    mediump float Rs_numerator = fresnel * geo * roughness;
    mediump float Rs_denominator  = dotNV * dotNL;
    mediump float Rs = Rs_numerator / Rs_denominator;
    
    mediump vec3 ambient = vec3( 0.1, 0.1, 0.1 );
    mediump vec3 final = clamp( max( 0.0, dotNL ) * (specularTex * Rs + albedoTex), 0.0, 1.0 );
    
    if (_LightCastsShadow == 1)
    {
        highp float depth = length( vPosition.xyz - _LightPositionWorld ) * uLinearDepthConstant;
        highp float shadow = VSM( depth );
        final *= shadow;
    }

    fragColor = vec4( ambient * albedoTex + final, 1.0 );
}

