cbuffer buffer : register(b0)
{
    column_major matrix _ViewProjectionMatrix;
    column_major matrix modelMatrix;
    column_major matrix _ViewMatrix;
#ifdef SKIN
    column_major matrix boneMatrices[8];
#endif
    column_major matrix _TPMatrix;
    float4 diffuseMap_ST;
    float3 _LightDirection;
    float3 _LightColor;
    float3 _LightPositionWorld;
    int _LightCastsShadow;
    float3 materialSpecular;
    float3 _CameraPosition;
    float roughness;
    float reflectivity;
    float uLinearDepthConstant;
}

struct VS_INPUT
{
    float3 Pos : POSITION;
    float2 Tex : TEXCOORD0;
    float3 Normal : NORMAL;
    float4 Tangent : TANGENT;
    float4 Color : COLOR;
#ifdef SKIN
    float4 BoneInd : BONE_INDEX;
    float4 Weights : BONE_WEIGHT;
#endif
};

struct PS_INPUT
{
    float4 Pos : SV_POSITION;
    float2 Tex : TEXCOORD0;
    float3 viewVec : VIEW_VEC;
    float3 lightVec : LIGHT_VEC;
#ifdef ENVIRONMENT_MAP
    float3 envTexCoord : ENV_TEXCOORD;
#endif
    float4 Position : POSITION;
    float4 ProjCoord : PROJ_COORD;
};

PS_INPUT VS( VS_INPUT input )
{
    PS_INPUT output = (PS_INPUT)0;
    float4 position = float4( input.Pos, 1.0f );
#ifdef SKIN
    int4 boneIndex = int4( input.BoneInd );
    
    position  = mul( boneMatrices[ boneIndex.x ], float4(input.Pos, 1.0f) * input.Weights.x );
    position += mul( boneMatrices[ boneIndex.y ], float4(input.Pos, 1.0f) * input.Weights.y );
    position += mul( boneMatrices[ boneIndex.z ], float4(input.Pos, 1.0f) * input.Weights.z );
    position += mul( boneMatrices[ boneIndex.w ], float4(input.Pos, 1.0f) * input.Weights.w );
#endif

    output.Pos = mul( _ViewProjectionMatrix, mul( modelMatrix, position ) );
    output.Tex = input.Tex * diffuseMap_ST.xy + diffuseMap_ST.zw;
    
    matrix modelViewMatrix = mul( _ViewMatrix, modelMatrix );
    float4 positionVS = mul( modelViewMatrix, position );

    float3x3 normalMatrix = (float3x3)( modelViewMatrix );

    float3 t = normalize( mul( normalMatrix, input.Tangent.xyz ) );
    float3 b = normalize( mul( normalMatrix, (cross( input.Normal, input.Tangent.xyz ) * input.Tangent.w)) );
    float3 n = normalize( mul( normalMatrix, input.Normal ) );
    float3x3 tbn = float3x3( t, b, n );

    output.lightVec = (mul( _ViewMatrix, float4( _LightDirection, 0.0f ) )).xyz;
    output.lightVec = mul( tbn, output.lightVec );
    output.viewVec = -positionVS.xyz;
    output.viewVec = mul( tbn, output.viewVec );

#ifdef ENVIRONMENT_MAP
    float4 worldPos = mul( modelMatrix, position );
    float4 worldNormal = normalize( mul( modelMatrix, float4(input.Normal, 0.0f) ) );
    float4 worldEye = normalize( worldPos - float4(_CameraPosition, 1.0f) );

    output.envTexCoord = reflect( worldEye.xyz, worldNormal.xyz );
#endif

    //output.Position = mul( modelMatrix, position );
    //output.ProjCoord = mul( _TPMatrix, position );

    output.Position = _LightCastsShadow == 1 ? mul( modelMatrix, position ) : float4( 0.0f, 0.0f, 0.0f, 0.0f );
    output.ProjCoord = _LightCastsShadow == 1 ? mul( _TPMatrix, position ) : float4(0.0f, 0.0f, 0.0f, 0.0f);

    return output;
}

Texture2D diffuseMap : register(t0);
Texture2D normalMap : register(t1);
TextureCube envMap : register(t2);
Texture2D specularMap : register(t3);
Texture2D _ShadowMap : register(t4);
SamplerState SampleTypeDiffuse : register(s0);
SamplerState SampleTypeNormal : register(s1);
#ifdef ENVIRONMENT_MAP
SamplerState SampleTypeEnv : register(s2);
#endif
#ifdef SPECULAR_MAP
SamplerState SampleTypeSpecular : register(s3);
#endif
SamplerState SampleTypeShadow : register(s4);

/*SamplerState SampleTypeShadow
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = CLAMP;
    AddressV = CLAMP;
    AddressW = CLAMP;
};*/

float linstep( float low, float high, float v )
{
    return saturate( (v - low) / (high - low) );
}

float VSM( float4 projCoord, float depth )
{
    float2 coord = projCoord.xy;
    coord.y = 1.0f - coord.y;

    float2 moments = _ShadowMap.Sample( SampleTypeShadow, coord ).rg;

    float variance = moments.y - moments.x * moments.x;

    float delta = depth - moments.x;
    float pMax = linstep( 0.2, 1.0, variance / (variance + delta * delta) );
    
    return (depth <= moments.x) ? 1.0f : pMax;
}

float4 PS( PS_INPUT input ) : SV_Target
{
    float3 albedoTex = diffuseMap.Sample( SampleTypeDiffuse, input.Tex ).rgb;
#ifdef SPECULAR_MAP
    float3 specularTex = specularMap.Sample( SampleTypeSpecular, input.Tex ).rgb;
#else
    float3 specularTex = materialSpecular;
#endif

    // input vectors
    float3 N = normalize( normalMap.Sample( SampleTypeNormal, input.Tex ).rgb * 2.0f - 1.0f );
    float3 V = normalize( input.viewVec );
    float3 L = normalize( input.lightVec );
    float3 H = normalize( L + V );

#ifdef ENVIRONMENT_MAP
    float3 envTex = envMap.Sample( SampleTypeEnv, input.envTexCoord + N.xyz ).rgb;
    albedoTex = lerp( albedoTex, envTex, reflectivity ) * specularTex; // Could use a reflection mask texture or channel.
#endif

    // dot products
    float dotNL = saturate( dot( N, L ) );
    float dotNH = saturate( dot( N, H ) );
    float dotNV = saturate( dot( N, V ) );
    float dotVH = saturate( dot( V, H ) );
    float r_sq = roughness * roughness;

    // Geometric term.
    float geoNumerator = 2.0f * dotNH;
    float geoB = (geoNumerator * dotNV) / dotVH;
    float geoC = (geoNumerator * dotNL) / dotVH;
    float geo = min( 1.0f, min( geoB, geoC ) );

    // Beckmann roughness.
    float roughness_a = 1.0f / (4.0f * r_sq * pow( dotNH, 4 ));
    float roughness_b = dotNH * dotNH - 1.0f;
    float roughness_c = r_sq * dotNH * dotNH;
    float roughness = roughness_a * exp( roughness_b / roughness_c );

    // Schlick Fresnel.
    float ref_at_norm_incidence = 1.0f;
    float fresnel = pow( 1.0f - dotVH, 5.0f );
    fresnel *= (1.0f - ref_at_norm_incidence);
    fresnel += ref_at_norm_incidence;

    // Specular BRDF.
    float Rs_numerator = fresnel * geo * roughness;
    float Rs_denominator = dotNV * dotNL;
    float Rs = Rs_numerator / Rs_denominator;

    float3 ambient = float3(0.05f, 0.05f, 0.05f);
    // FIXME: Should not need saturate() here, probably something is not normalized earlier.
    float3 final = ambient * albedoTex + saturate(max( 0.0f, dotNL ) * (specularTex * Rs + albedoTex));

    float depth = length( input.Position.xyz - _LightPositionWorld ) * uLinearDepthConstant;
    float shadow = (_LightCastsShadow == 1) ? VSM( input.ProjCoord, depth ) : 1.0f;

    return float4( final * shadow, 1.0f );
}