#include "inspector.hpp"
#include <QEvent>
#include <QFileDialog>
#include <QTableWidget>
#include <QBoxLayout>
#include <QWidget>
#include <QLineEdit>
#include <iostream>
#include "Aether3D_Include/Aether3D.hpp"
#include "EditorState.h"

// TODO: Own inspector class for every property.
struct Inspector::Impl
{
    QLineEdit* nodeNameEdit = nullptr;
    QTableWidget* meshListTable = nullptr;
    QTableWidget* transformTable = nullptr;
    QTableWidget* lightTable = nullptr;
    QWidget* widget = nullptr;
};

Inspector::Inspector( Aether3D_API* aDLL, EditorState* state )
    : m( new Impl )
    , dll( aDLL )
    , editorState( state )
{
}

Inspector::~Inspector()
{
}

QWidget* Inspector::GetWidget()
{
    return m->widget;
}

QLineEdit* Inspector::GetNodeNameEdit()
{
    return m->nodeNameEdit;
}

QTableWidget* Inspector::GetTransformTable()
{
    return m->transformTable;
}

QTableWidget* Inspector::GetLightTable()
{
    return m->lightTable;
}

void Inspector::Init()
{
    m->nodeNameEdit = new QLineEdit();
    m->nodeNameEdit->setFixedHeight( 25 );

    m->transformTable = new QTableWidget( 3, 3 );
    m->transformTable->setHorizontalHeaderLabels(QString("X;Y;Z").split(";"));
    m->transformTable->setVerticalHeaderLabels(QString("Position;Rotation;Scale").split(";"));
    m->transformTable->setSpan( 2, 0, 1, 3 );

    m->meshListTable = new QTableWidget( 2, 1 );
    m->meshListTable->setHorizontalHeaderLabels(QString("Material").split(";"));
    m->meshListTable->setEditTriggers( QAbstractItemView::NoEditTriggers );

    m->lightTable = new QTableWidget( 2, 3 );
    m->lightTable->setHorizontalHeaderLabels(QString("X/R;Y/G;Z/B").split(";"));
    m->lightTable->setVerticalHeaderLabels(QString("Direction;Color;Scale").split(";"));

    QBoxLayout* inspectorLayout = new QBoxLayout( QBoxLayout::TopToBottom );
    inspectorLayout->addWidget( m->nodeNameEdit );
    inspectorLayout->addWidget( m->transformTable );
    inspectorLayout->addWidget( m->meshListTable );
    inspectorLayout->addWidget( m->lightTable );
    m->widget = new QWidget();
    m->widget->setLayout( inspectorLayout );

    connect(m->meshListTable, SIGNAL(cellClicked(int,int)), this, SLOT(MeshCellClicked(int,int)));
}

void Inspector::MeshCellClicked( int row, int )
{
    if (editorState->selectedNode && editorState->selectedNode->GetType() == SceneGraphNode::Type::Model)
    {
        const std::string path = QFileDialog::getOpenFileName( this, "Open Material", "", "Materials (*.material)" ).toStdString();

        if (path.empty())
        {
            return;
        }

        Material* mat = dll->MakeMaterial();
        mat->LoadFromFile( path.c_str() );

        Model* model = static_cast<Model*>(editorState->selectedNode);
        model->SetMaterialForMesh( mat, m->meshListTable->verticalHeaderItem( row )->text().toUtf8().constData() );
        UpdateModel( model );
    }
}

void Inspector::SetNodeName( const char* name )
{
    m->nodeNameEdit->setText( name );
}

const char* Inspector::GetNodeName() const
{
    return m->nodeNameEdit->text().toUtf8().constData();
}

void Inspector::UpdateModel( const Model* model )
{
    auto meshNames = model->GetMeshNames();
    const int meshCount = meshNames.size();
    m->meshListTable->setRowCount( meshCount );

    int row = 0;

    for (auto meshName : meshNames)
    {
        auto headerItem = new QTableWidgetItem( meshName.c_str() );
        m->meshListTable->setVerticalHeaderItem( row, headerItem );

        std::string materialName = "material";
        auto material = const_cast<Model*>(model)->GetMeshMaterial(meshName.c_str());

        if (material)
        {
            materialName = material->Name();
        }

        auto materialItem = new QTableWidgetItem( materialName.c_str() );
        m->meshListTable->setItem( row, 0, materialItem );
        ++row;
    }
}

void Inspector::UpdateLight( const Light* light )
{
    // Color.
    const Vec3 color = light->GetColor();

    char buf[ 50 ];
    sprintf( buf, "%.2f", color.x );
    m->lightTable->setItem( 1, 0, new QTableWidgetItem( buf ) );

    sprintf( buf, "%.2f", color.y );
    m->lightTable->setItem( 1, 1, new QTableWidgetItem( buf ) );

    sprintf( buf, "%.2f", color.z );
    m->lightTable->setItem( 1, 2, new QTableWidgetItem( buf ) );
}

void Inspector::Update( const SceneGraphNode* selectedNode )
{
    if (selectedNode)
    {
        const bool isModel = selectedNode->GetType() == SceneGraphNode::Type::Model;
        m->meshListTable->setVisible( isModel );

        if (isModel)
        {
            m->meshListTable->show();
            UpdateModel( static_cast< const Model* >( selectedNode ) );
        }
        else
        {
            m->meshListTable->hide();
        }

        const bool isLight = selectedNode->GetType() == SceneGraphNode::Type::Light;

        if (isLight)
        {
            m->lightTable->show();
            UpdateLight( static_cast< const Light* >( selectedNode ) );
        }
        else
        {
            m->lightTable->hide();
        }

        // Position.
        const Vec3 position = selectedNode->GetPosition();

        char buf[ 50 ];
        sprintf( buf, "%.2f", position.x );
        m->transformTable->setItem( 0, 0, new QTableWidgetItem( buf ) );

        sprintf( buf, "%.2f", position.y );
        m->transformTable->setItem( 0, 1, new QTableWidgetItem( buf ) );

        sprintf( buf, "%.2f", position.z );
        m->transformTable->setItem( 0, 2, new QTableWidgetItem( buf ) );

        // Rotation.
        const Quaternion rotation = selectedNode->GetRotation();
        const Vec3 euler = rotation.GetEuler();

        //std::cout << "Rotation quaternion: " << rotation.x << " " << rotation.y << " " << rotation.z << " " << rotation.w << std::endl;
        //std::cout << "Rotation euler: " << euler.x << " " << euler.y << " " << euler.z << " " << std::endl;

        sprintf( buf, "%.2f", euler.x );
        m->transformTable->setItem( 1, 0, new QTableWidgetItem( buf ) );

        sprintf( buf, "%.2f", euler.y );
        m->transformTable->setItem( 1, 1, new QTableWidgetItem( buf ) );

        sprintf( buf, "%.2f", euler.z );
        m->transformTable->setItem( 1, 2, new QTableWidgetItem( buf ) );

        // Scale.
        sprintf( buf, "%.2f", selectedNode->GetScale() );
        m->transformTable->setItem( 2, 0, new QTableWidgetItem( buf ) );
    }
    else
    {
        m->transformTable->setItem( 0, 0, new QTableWidgetItem( "" ) );
        m->transformTable->setItem( 0, 1, new QTableWidgetItem( "" ) );
        m->transformTable->setItem( 0, 2, new QTableWidgetItem( "" ) );

        m->transformTable->setItem( 1, 0, new QTableWidgetItem( "" ) );
        m->transformTable->setItem( 1, 1, new QTableWidgetItem( "" ) );
        m->transformTable->setItem( 1, 2, new QTableWidgetItem( "" ) );

        m->transformTable->setItem( 2, 0, new QTableWidgetItem( "" ) );
    }
}
