#include "mainwindow.h"
#include <iostream>
#include <fstream>
#include <cassert>
#include <cstdio>
#include <list>
#include "myglwidget.hpp"
#include "inspector.hpp"
#include <QtGui>
#include <QAction>
#include <QFileDialog>
#include <QLabel>
#include <QList>
#include <QLineEdit>
#include <QTableWidget>
#include <QTableWidgetItem>
#include <QTreeView>
#include <QTreeWidget>
#include <QListWidgetItem>
#include <QListWidget>
#include <QIcon>
#include <QMenu>
#include <QMenuBar>
#include <QSplitter>
#include <QWindow>
#include <QtOpenGL/QGLFormat>
#include "Aether3D_Include/Aether3D.hpp"
#include "Aether3D_Include/Quaternion.hpp"
#include "EditorState.h"

std::string AbsoluteFilePath( const std::string& relativePath )
{
    QDir dir = QDir::currentPath();
#if __APPLE__
    dir.cdUp();
    dir.cdUp();
    dir.cdUp();
#endif
    return dir.absoluteFilePath( relativePath.c_str() ).toUtf8().constData();
}

class Menu
{
public:
    QMenu* sceneMenu = nullptr;
    QMenuBar* menuBar = nullptr;
    QAction* createCameraAction = nullptr;
    QAction* createPointLightAction = nullptr;
    QAction* createDirectionalLightAction = nullptr;
    QAction* createEmptyAction = nullptr;
    QAction* openModelAction = nullptr;
    QMenu* fileMenu = nullptr;
    QAction* saveSceneAction = nullptr;
    QAction* loadSceneAction = nullptr;
    QMenu* recentScenesMenu = nullptr;
    std::list< QAction* > recentScenesActions;
};

struct MainWindow::Impl
{
    Impl( Aether3D_API* dll, EditorState* state )
        : inspector( dll, state )
    {}

    void UpdateRecentsMenu();

    std::map< QTreeWidgetItem*, SceneGraphNode* > treeWidgetToNode;
    Menu menu;
    Inspector inspector;
    QListWidget* assetList = nullptr;
    QIcon* iconMaterial = nullptr;
    QWidget* window = nullptr;
};

void MainWindow::Impl::UpdateRecentsMenu()
{
    menu.recentScenesMenu = menu.fileMenu->addMenu( tr("&Recent Scenes") );
    QDir homeDir = QDir::home();
    std::string recentsFile = homeDir.absoluteFilePath( ".recent_scenes" ).toUtf8().constData();
    std::ifstream ifs( recentsFile.c_str() );

    while (ifs && !ifs.eof())
    {
        std::string line;
        getline( ifs, line );

        std::ifstream ifs( line );

        if (ifs)
        {
            menu.recentScenesActions.push_back( menu.recentScenesMenu->addAction( line.c_str() ) );
        }
    }
}

MainWindow::MainWindow( QApplication* app, Aether3D_API* dll, EditorState* state, QWidget* )
    : m( new Impl( dll, state ) )
    , editorState( state )
{
    QGLFormat gFormat;
    gFormat.setVersion( 3, 2 );
    gFormat.setSwapInterval( 1 );
    gFormat.setProfile( QGLFormat::CoreProfile );

    sceneWidget = new MyGLWidget( gFormat, app, dll, this, editorState );

    InitMenu();

    m->iconMaterial = new QIcon( AbsoluteFilePath( "glider.png").c_str() );

    m->assetList = new QListWidget( this );
    QListWidgetItem* newItem = new QListWidgetItem;
    newItem->setText( "test.material" );
    newItem->setIcon( *m->iconMaterial );
    m->assetList->insertItem( 0, newItem );

    sceneTree = new QTreeWidget();
    sceneTree->setColumnCount( 1 );
    connect( sceneTree, &QTreeWidget::itemClicked, [&](QTreeWidgetItem* item, int /* column */) { SelectTreeItem( item ); });

    m->inspector.Init();
    connect( m->inspector.GetNodeNameEdit(), &QLineEdit::editingFinished, [&] { NodeNameChanged(); } );
    connect( m->inspector.GetTransformTable(), &QTableWidget::itemChanged, [&](QTableWidgetItem * item) { InspectorTransformChanged( item ); } );
    connect( m->inspector.GetLightTable(), &QTableWidget::itemChanged, [&](QTableWidgetItem * item) { InspectorLightChanged( item ); } );

    QSplitter* splitter = new QSplitter();
    splitter->addWidget( sceneTree );
    splitter->addWidget( sceneWidget );
    splitter->addWidget( m->inspector.GetWidget() );

    QSplitter* splitter2 = new QSplitter( Qt::Orientation::Vertical );
    splitter2->addWidget( splitter );
    splitter2->addWidget( m->assetList );

    setCentralWidget( splitter2 );
    setWindowTitle( tr( "Editor" ) );
}

MainWindow::~MainWindow()
{
    /*delete sceneTree;
    delete sceneWidget;
    delete assetList;
    delete menuBar;
    delete createPointLightAction;
    delete createDirectionalLightAction;
    delete openModelAction;
    delete transformTable;*/
    //std::cout << "Destroying mainwindow." << std::endl;
    sceneWidget->DeinitRenderer();
}

void MainWindow::InitMenu()
{    
    m->menu.createPointLightAction = new QAction( this );
    m->menu.createPointLightAction->setText( tr("Create Point Light") );
    connect( m->menu.createPointLightAction, SIGNAL(triggered()), this, SLOT(CreatePointLight()) );

    m->menu.createDirectionalLightAction = new QAction( this );
    m->menu.createDirectionalLightAction->setText( tr("Create Directional Light") );
    connect( m->menu.createDirectionalLightAction, SIGNAL( triggered() ), this, SLOT( CreateDirectionalLight() ) );

    m->menu.createEmptyAction = new QAction( this );
    m->menu.createEmptyAction->setText( tr("Create Empty Node") );
    connect( m->menu.createEmptyAction, SIGNAL( triggered() ), this, SLOT( CreateEmptyNode() ) );

    m->menu.createCameraAction = new QAction( this );
    m->menu.createCameraAction->setText( tr("Create Camera") );
    connect( m->menu.createCameraAction, SIGNAL( triggered() ), this, SLOT( CreateCamera() ) );

    m->menu.openModelAction = new QAction( this );
    m->menu.openModelAction->setText( tr("Open Model") );
    connect( m->menu.openModelAction, SIGNAL( triggered() ), this, SLOT( OpenModel() ) );

    m->menu.saveSceneAction = new QAction( this );
    m->menu.saveSceneAction->setText( tr("Save Scene") );
    connect( m->menu.saveSceneAction, SIGNAL( triggered() ), this, SLOT( SaveScene() ) );

    m->menu.loadSceneAction = new QAction( this );
    m->menu.loadSceneAction->setText( tr("Load Scene") );
    connect( m->menu.loadSceneAction, SIGNAL( triggered() ), this, SLOT( OpenScene() ) );

    m->menu.menuBar = new QMenuBar( this );
    m->menu.sceneMenu = m->menu.menuBar->addMenu( tr("&SceneNode") );
    m->menu.sceneMenu->addAction( m->menu.createPointLightAction );
    m->menu.sceneMenu->addAction( m->menu.createDirectionalLightAction );
    m->menu.sceneMenu->addAction( m->menu.createEmptyAction );
    m->menu.sceneMenu->addAction( m->menu.createCameraAction );
    m->menu.sceneMenu->addAction( m->menu.openModelAction );

    m->menu.fileMenu = m->menu.menuBar->addMenu( tr("&File") );
    m->menu.fileMenu->addAction( m->menu.loadSceneAction );
    m->menu.fileMenu->addAction( m->menu.saveSceneAction );

    m->UpdateRecentsMenu();

    connect( m->menu.recentScenesMenu, &QMenu::triggered, this, &MainWindow::OpenRecentScene );

    setMenuBar( m->menu.menuBar );
}

void MainWindow::OpenRecentScene(QAction* action)
{
    std::string str( action->text().toStdString() );
    std::ifstream ifs( str.c_str() );

    if (!ifs)
    {
        std::cerr << str.c_str() << " not found!" << std::endl;
        return;
    }

    sceneWidget->PreSave();
    sceneWidget->GetScene()->LoadFromFile( str.c_str() );

    Camera* playerCamera = sceneWidget->GetScene()->GetNonRTCamera();

    if (playerCamera && playerCamera != sceneWidget->GetEditorCamera())
    {
        playerCamera->SetRenderTarget( sceneWidget->CameraRenderTexture() );
    }

    sceneWidget->PostSave();

    UpdateHierarchy();
    setWindowTitle( str.c_str() );
}

void MainWindow::closeEvent(QCloseEvent *)
{
    std::cout << "Closing mainwindow." << std::endl;
}

void MainWindow::NodeNameChanged()
{
    if (editorState->selectedNode)
    {
        editorState->selectedNode->SetName( m->inspector.GetNodeName() );
    }

    UpdateHierarchy();
}

void MainWindow::InspectorLightChanged( QTableWidgetItem* item )
{
    assert( editorState );

    if (!editorState->selectedNode || editorState->selectedNode->GetType() != SceneGraphNode::Type::Light)
    {
        return;
    }

    Light* light = static_cast< Light* >( editorState->selectedNode );
    Vec3 color = light->GetColor();
    const std::string newValue = item->text().toUtf8().constData();

    if (item->row() == 1 && item->column() == 0)
    {
        try
        {
            color.x = std::stof( newValue );
        }
        catch (std::invalid_argument& a)
        {
            color.x = 0;
            item->setText( "0" );
        }
    }
    else if (item->row() == 1 && item->column() == 1)
    {
        try
        {
            color.y = std::stof( newValue );
        }
        catch (std::invalid_argument& a)
        {
            color.y = 0;
            item->setText( "0" );
        }
    }
    else if (item->row() == 1 && item->column() == 2)
    {
        try
        {
            color.z = std::stof( newValue );
        }
        catch (std::invalid_argument& a)
        {
            color.z = 0;
            item->setText( "0" );
        }
    }

    light->SetColor( color );
}

void MainWindow::InspectorTransformChanged( QTableWidgetItem* item )
{
    assert( editorState );

    if (!editorState->selectedNode)
    {
        return;
    }

    Vec3 position = editorState->selectedNode->GetPosition();
    Quaternion rotation = editorState->selectedNode->GetRotation();
    Vec3 rotationEuler = rotation.GetEuler();

    const std::string newValue = item->text().toUtf8().constData();

    if (item->row() == 0 && item->column() == 0)
    {
        try
        {
            position.x = std::stof( newValue );
        }
        catch (std::invalid_argument& a)
        {
            position.x = 0;
            item->setText( "0" );
        }
    }
    else if (item->row() == 0 && item->column() == 1)
    {
        try
        {
            position.y = std::stof( newValue );
        }
        catch (std::invalid_argument& a)
        {
            position.y = 0;
            item->setText( "0" );
        }
    }
    else if (item->row() == 0 && item->column() == 2)
    {
        try
        {
            position.z = std::stof( newValue );
        }
        catch (std::invalid_argument& a)
        {
            position.z = 0;
            item->setText( "0" );
        }
    }
    else if (item->row() == 1 && item->column() == 0) // Rotation x
    {
        try
        {
            rotationEuler.x = std::stof( newValue );
        }
        catch (std::invalid_argument& a)
        {
            std::cout << "Invalid rot x" << std::endl;
            rotationEuler.x = 0;
            item->setText( "0" );
        }
    }
    else if (item->row() == 1 && item->column() == 1) // Rotation y
    {
        try
        {
            rotationEuler.y = std::stof( newValue );
        }
        catch (std::invalid_argument& a)
        {
            std::cout << "Invalid rot y" << std::endl;
            rotationEuler.y = 0;
            item->setText( "0" );
        }
    }
    else if (item->row() == 1 && item->column() == 2) // Rotation z
    {
        try
        {
            rotationEuler.z = std::stof( newValue );
        }
        catch (std::invalid_argument& a)
        {
            std::cout << "Invalid rot z" << std::endl;
            rotationEuler.z = 0;
            item->setText( "0" );
        }
    }
    else if (item->row() == 2 && item->column() == 0)
    {
        try
        {
            const float scale = std::stof( newValue );
            editorState->selectedNode->SetScale( scale );
        }
        catch (std::invalid_argument& a)
        {
            item->setText( std::to_string(editorState->selectedNode->GetScale()).c_str());
        }
    }

    rotation.FromEuler( rotationEuler );

    //std::cout << "Setting rotation from Euler: " << rotationEuler.x << ", " << rotationEuler.y << ", " << rotationEuler.z << std::endl;

    editorState->selectedNode->SetPosition( position );
    editorState->selectedNode->SetRotation( rotation );
    sceneWidget->SelectionPositionChanged( position );
}

void MainWindow::SelectTreeItem( QTreeWidgetItem* item )
{
    SetSelectedNode( m->treeWidgetToNode[ item ] );
}

void MainWindow::RemoveNode( SceneGraphNode* node )
{
    sceneWidget->RemoveNode( node );
    UpdateHierarchy();
    UpdateInspector();
}

void MainWindow::keyPressEvent( QKeyEvent* aEvent )
{
    const int macDelete = 16777219;

    if (aEvent->key() == Qt::Key_Escape)
    {
        std::cout << "MainWindow: Esc" << std::endl;
    }
    else if (aEvent->key() == Qt::Key_Delete || aEvent->key() == macDelete)
    {
        std::cout << "MainWindow: Delete" << std::endl;
        RemoveNode( editorState->selectedNode );
    }
    else if (aEvent->key() == Qt::Key_Enter)
    {
        std::cout << "MainWindow: Enter" << std::endl;
    }
    else if (aEvent->key() == Qt::Key_F && editorState->selectedNode)
    {
        sceneWidget->CenterSelected();
    }
    else if (aEvent->key() == Qt::Key_S && aEvent->modifiers() & Qt::ControlModifier)
    {
        SaveScene();
    }
}

void MainWindow::Save()
{
    SaveScene();
}

void MainWindow::UpdateHierarchy()
{
    m->treeWidgetToNode.clear();
    sceneTree->clear();
    auto root = sceneWidget->RootNode();
    QList<QTreeWidgetItem *> nodes;

    for (auto& node : root->Children())
    {
        if (!sceneWidget->IsEditorWidget( node ))
        {
            nodes.append(new QTreeWidgetItem((QTreeWidget*)0, QStringList(QString(node->GetName().c_str()))));
            m->treeWidgetToNode[ nodes.back() ] = node;

            if (node == editorState->selectedNode)
            {
                sceneTree->setCurrentItem( nodes.back() );
            }
        }
    }

    sceneTree->insertTopLevelItems( 0, nodes );
    sceneTree->setHeaderLabel( tr("Hierarchy") );

    if (editorState->selectedNode)
    {
        int i = 0;

        for (auto& node : root->Children())
        {
            if (!sceneWidget->IsEditorWidget( node ))
            {
                if (node == editorState->selectedNode)
                {
                    sceneTree->setCurrentItem( nodes.at( i ) );
                }

                ++i;
            }
        }
    }
    else
    {
        sceneTree->clearSelection();
    }
    /*
    QList<QTreeWidgetItem *> items;
    for (int i = 0; i < 10; ++i)
        items.append(new QTreeWidgetItem((QTreeWidget*)0, QStringList(QString("item: %1").arg(i))));

    items[ 8 ]->addChild( new QTreeWidgetItem((QTreeWidget*)0, QStringList(QString("inner"))));
    sceneView->insertTopLevelItems( 0, items );
    sceneView->setHeaderLabel( tr("Hierarchy") );
    */
}

void MainWindow::CreatePointLight()
{
    sceneWidget->AddLight( Light::Type::Point );
    UpdateHierarchy();
}

void MainWindow::CreateDirectionalLight()
{
    sceneWidget->AddLight( Light::Type::Directional );
    UpdateHierarchy();
}

void MainWindow::CreateEmptyNode()
{
    sceneWidget->AddEmpty();
    UpdateHierarchy();
}

void MainWindow::CreateCamera()
{
    sceneWidget->AddCamera();
    UpdateHierarchy();
}

void MainWindow::OpenModel()
{    
    QString fileName = QFileDialog::getOpenFileName( this, "Open Model", "", "Models (*.ae3d)" );

    if (fileName != "")
    {
        sceneWidget->AddModel( fileName.toStdString().c_str() );
        UpdateHierarchy();
    }
    else
    {
        std::cout << "Operation cancelled." << std::endl;
    }
}

void MainWindow::OpenScene()
{
#if __APPLE__
    const char* appendDir = "/Documents";
#else
    const char* appendDir = "";
#endif

    QString fileName = QFileDialog::getOpenFileName( this, "Open Scene", QDir::homePath() + appendDir, "Scenes (*.scene)" );

    const bool success = sceneWidget->GetScene()->LoadFromFile( fileName.toStdString() );

    if (!success)
    {
        std::cerr << "Could not open " << fileName.toStdString() << std::endl;
        return;
    }

    Camera* playerCamera = sceneWidget->GetScene()->GetNonRTCamera();

    if (playerCamera && playerCamera != sceneWidget->GetEditorCamera())
    {
        playerCamera->SetRenderTarget( sceneWidget->CameraRenderTexture() );
    }

    setWindowTitle( fileName );
    UpdateHierarchy();

    QDir homeDir = QDir::home();
    std::string recentsFile = homeDir.absoluteFilePath( ".recent_scenes" ).toUtf8().constData();
    std::ofstream ofs( recentsFile.c_str(), std::ios::app );
    ofs << fileName.toUtf8().constData() << std::endl;
}

void MainWindow::SaveScene()
{
    sceneWidget->PreSave();

#if __APPLE__
    const char* appendDir = "/Documents";
#else
    const char* appendDir = "";
#endif

    QString fileName = QFileDialog::getSaveFileName( this, "Save Scene", QDir::homePath() + appendDir, "Scenes (*.scene)" );
    const bool saveSuccess = sceneWidget->GetScene()->SaveToFile( fileName.toUtf8().constData() );

    if (!saveSuccess)
    {
        std::cout << "Could not save scene into " << fileName.toUtf8().constData() << std::endl;
    }

    sceneWidget->PostSave();
}

void MainWindow::SetSelectedNode( SceneGraphNode* node )
{
    if (node)
    {
        std::cout << "selected: " << node->GetName() << std::endl;
    }

    editorState->selectedNode = node;
    UpdateHierarchy();
    UpdateInspector();
    sceneWidget->SelectionChanged();
    m->inspector.SetNodeName( node ? node->GetName().c_str() : "" );
}

void MainWindow::UpdateInspector()
{
    m->inspector.Update( editorState->selectedNode );
}
