#include "myglwidget.hpp"
#include "mainwindow.h"
#include <QApplication>
#include <QDir>
#include <QKeyEvent>
#include <algorithm>
#include <iostream>
#include <list>
#include <vector>
#include <cassert>
#include "Aether3D_Include/Aether3D.hpp"
#include "Aether3D_Include/MathUtil.hpp"
#include "EditorState.h"

std::string AbsoluteFilePath( const std::string& relativePath );

void GetSceneNodesByType( SceneGraphNode* node, SceneGraphNode::Type type, std::list< SceneGraphNode* >& outNodes )
{
    if (node->GetType() == type)
    {
        outNodes.push_back( node );
    }

    for (auto child : node->Children())
    {
        GetSceneNodesByType( child, type, outNodes );
    }
}

MyGLWidget::MyGLWidget(QWidget *parent)
    : QGLWidget(parent)
{
}

MyGLWidget::MyGLWidget(const QGLFormat & format, QApplication* aApp, Aether3D_API* aDLL, MainWindow* aMainWindow, EditorState* state, QWidget * parent)
    : QGLWidget( format, parent )
    , dll( aDLL )
    , app( aApp )
    , mainWindow( aMainWindow )
    , editorState( state )
{
}

void MyGLWidget::initializeGL()
{
    renderer = dll->MakeRenderer( AbsoluteFilePath("builtin") );
    renderer->OpenWindow( 200, 200, false, 1, false );
    renderer->SetClearColor( 0, 0, 0 );

    editorCamera = dll->MakeCamera();
    editorCamera->SetProjection( 55, renderer->Width() / (float)renderer->Height(), 1, 200 );
    editorCamera->LookAt( Vec3( 0, 0, 5 ), Vec3( 0, 0, 0 ), Vec3( 0, 1, 0 ) );
    editorCamera->SetName( "editor camera" );

    scene = dll->MakeScene();
    scene->Add( editorCamera );

    Error::Enum error = Error::NO_ERRORS;
    sphere = dll->MakeModel( AbsoluteFilePath( "builtin/editor/models/sphere.ae3d" ), error );
    sphere->SetName( "sphere" );
    scene->Add( sphere );

    cursorTranslate = dll->MakeModel( AbsoluteFilePath( "builtin/editor/models/cursor_translate.ae3d" ), error );
    cursorTranslate->SetName( "cursorTranslate" );
    scene->Add( cursorTranslate );

    checker = dll->MakeTexture( AbsoluteFilePath( "builtin/textures/checker.png" ), true, true, Texture::Tiling::Repeat, Texture::Filter::Linear );
    lightTex = dll->MakeTexture( AbsoluteFilePath( "builtin/textures/light.png" ), true, true, Texture::Tiling::Repeat, Texture::Filter::Linear );
    cameraTex = dll->MakeTexture( AbsoluteFilePath( "builtin/textures/camera.png" ), true, true, Texture::Tiling::Repeat, Texture::Filter::Linear );
    cameraRenderTexture = dll->MakeRenderTexture(320, 200);

    unlit = dll->MakeMaterial();
    unlit->SetShaders( "unlit", "unlit", "", "" );
    unlit->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    unlit->SetTexture( "textureMap", checker );

    sphere->SetMaterialForMesh( unlit, "Sphere" );

    app->installEventFilter( this );
    connect(&myTimer, SIGNAL( timeout() ), this, SLOT( updateCamera() ));
    myTimer.start();

    InitGizmoMaterials();
}

void MyGLWidget::DrawSelectionHighlight()
{
    if (editorState->selectedNode && editorState->selectedNode->GetType() == SceneGraphNode::Type::Model)
    {
        Model* model = static_cast< Model* >( editorState->selectedNode );
        Vec3 aabbMin, aabbMax;
        model->GetAABB( aabbMin, aabbMax );
        std::vector< Vec3 > corners;
        MathUtil::GetCorners( aabbMin, aabbMax, corners );
        std::vector< Vec3 > lines( 24 );

        lines[0 ] = corners[ 0 ];
        lines[1 ] = corners[ 1 ];

        lines[2 ] = corners[ 0 ];
        lines[3 ] = corners[ 2 ];

        lines[4 ] = corners[ 0 ];
        lines[5 ] = corners[ 3 ];

        lines[6 ] = corners[ 3 ];
        lines[7 ] = corners[ 5 ];

        lines[8 ] = corners[ 2 ];
        lines[9 ] = corners[ 4 ];

        lines[10] = corners[ 5 ];
        lines[11] = corners[ 6 ];

        lines[12] = corners[ 6 ];
        lines[13] = corners[ 7 ];

        lines[14] = corners[ 4 ];
        lines[15] = corners[ 6 ];

        lines[16] = corners[ 3 ];
        lines[17] = corners[ 7 ];

        lines[18] = corners[ 1 ];
        lines[19] = corners[ 7 ];

        lines[20] = corners[ 1 ];
        lines[21] = corners[ 4 ];

        lines[22] = corners[ 2 ];
        lines[23] = corners[ 5 ];

        renderer->DrawLines( lines.data(), (int)lines.size(), Vec4( 1, 1, 1, 1 ) );
    }
}

bool MyGLWidget::IsEditorWidget( const SceneGraphNode* node ) const
{
    return node == editorCamera || node == cursorTranslate;
}

void MyGLWidget::updateCamera()
{
    const float speed = 0.2f;
    editorCamera->MoveRight( cameraMoveDir.x * speed );
    editorCamera->MoveUp( cameraMoveDir.y * speed );
    editorCamera->MoveForward( cameraMoveDir.z * speed );
    updateGL();
}

void MyGLWidget::updateGL()
{
    update();
}

void MyGLWidget::paintGL()
{
    renderer->ClearScreen();
    renderer->Draw( scene );
    DrawLightSprites();
    DrawCameraSprites();
    DrawSelectionHighlight();

    if (editorState->selectedNode && editorState->selectedNode->GetType() == SceneGraphNode::Type::Camera)
    {
        renderer->Draw(cameraRenderTexture, 0, cameraRenderTexture->Height(), cameraRenderTexture->Width(), -cameraRenderTexture->Height(), Vec4(1, 1, 1, 1));
    }
    renderer->FinishFrame();
}

void MyGLWidget::resizeGL( int width, int height )
{
    renderer->Resize( width, height );
    editorCamera->SetProjection( 55, renderer->Width() / (float)renderer->Height(), 1, 200 );
}

void MyGLWidget::keyPressEvent( QKeyEvent* aEvent )
{
    if (aEvent->key() == Qt::Key_Escape)
    {
        mainWindow->SetSelectedNode( nullptr );
    }
    else if (aEvent->key() == Qt::Key_A && grabMouse)
    {
        cameraMoveDir.x = -1;
    }
    else if (aEvent->key() == Qt::Key_D && grabMouse)
    {
        cameraMoveDir.x = 1;
    }
    else if (aEvent->key() == Qt::Key_Q && grabMouse)
    {
        cameraMoveDir.y = -1;
    }
    else if (aEvent->key() == Qt::Key_E && grabMouse)
    {
        cameraMoveDir.y = 1;
    }
    else if (aEvent->key() == Qt::Key_W && grabMouse)
    {
        cameraMoveDir.z = -1;
    }
    else if (aEvent->key() == Qt::Key_S && grabMouse)
    {
        cameraMoveDir.z = 1;
    }
    else if (aEvent->key() == Qt::Key_D && aEvent->modifiers() & Qt::ControlModifier)
    {
        if (editorState->selectedNode && editorState->selectedNode->GetType() == SceneGraphNode::Type::Model)
        {
            auto path = static_cast< Model* >( editorState->selectedNode )->GetPath();
            Model* model = AddModel( path.c_str() );

            if (model)
            {
                model->SetPosition( editorState->selectedNode->GetPosition() + Vec3( 0.5f, 0.5f, 0 ) );
                mainWindow->SetSelectedNode( model );
                mainWindow->UpdateHierarchy();
            }
        }
        else if (editorState->selectedNode && editorState->selectedNode->GetType() == SceneGraphNode::Type::Light)
        {
            const auto selectedLight = static_cast< Light* >( editorState->selectedNode );
            Light* light = AddLight( selectedLight->GetLightType() );
            light->SetPosition( selectedLight->GetPosition() + Vec3( 0.5f, 0.5f, 0 ) );
            light->SetColor( selectedLight->GetColor() );
            mainWindow->SetSelectedNode( light );
            mainWindow->UpdateHierarchy();
        }
    }
    else if (aEvent->key() == Qt::Key_S && aEvent->modifiers() & Qt::ControlModifier)
    {
        mainWindow->Save();
    }
}

void MyGLWidget::keyReleaseEvent( QKeyEvent* aEvent )
{
    const int macDelete = 16777219;

    if (aEvent->key() == Qt::Key_Delete || aEvent->key() == macDelete)
    {
        mainWindow->RemoveNode( editorState->selectedNode );
    }
    else if (aEvent->key() == Qt::Key_A)
    {
        cameraMoveDir.x = 0;
    }
    else if (aEvent->key() == Qt::Key_D)
    {
        cameraMoveDir.x = 0;
    }
    else if (aEvent->key() == Qt::Key_Q)
    {
        cameraMoveDir.y = 0;
    }
    else if (aEvent->key() == Qt::Key_E)
    {
        cameraMoveDir.y = 0;
    }
    else if (aEvent->key() == Qt::Key_F)
    {
        if (editorState->selectedNode)
        {
            CenterSelected();
        }
    }
    else if (aEvent->key() == Qt::Key_W)
    {
        cameraMoveDir.z = 0;
    }
    else if (aEvent->key() == Qt::Key_S)
    {
        cameraMoveDir.z = 0;
    }
}

void MyGLWidget::InitGizmoMaterials()
{
    gizmoMaterials.xTexture = dll->MakeTexture( AbsoluteFilePath( "builtin/editor/textures/xaxis.png" ), true, true, Texture::Tiling::Repeat, Texture::Filter::Linear );
    gizmoMaterials.yTexture = dll->MakeTexture( AbsoluteFilePath( "builtin/editor/textures/yaxis.png" ), true, true, Texture::Tiling::Repeat, Texture::Filter::Linear );
    gizmoMaterials.zTexture = dll->MakeTexture( AbsoluteFilePath( "builtin/editor/textures/zaxis.png" ), true, true, Texture::Tiling::Repeat, Texture::Filter::Linear );

    gizmoMaterials.xAxis = dll->MakeMaterial();
    gizmoMaterials.xAxis->SetShaders( "unlit", "unlit", "", "" );
    gizmoMaterials.xAxis->SetTexture("textureMap", gizmoMaterials.xTexture );
    gizmoMaterials.xAxis->SetVector("tint", Vec4( 0.2f, 0.2f, 0.2f, 1 ) );

    gizmoMaterials.yAxis = dll->MakeMaterial();
    gizmoMaterials.yAxis->SetShaders( "unlit", "unlit", "", "" );
    gizmoMaterials.yAxis->SetTexture("textureMap", gizmoMaterials.yTexture );
    gizmoMaterials.yAxis->SetVector("tint", Vec4( 0.2f, 0.2f, 0.2f, 1 ) );

    gizmoMaterials.zAxis = dll->MakeMaterial();
    gizmoMaterials.zAxis->SetShaders( "unlit", "unlit", "", "" );
    gizmoMaterials.zAxis->SetTexture("textureMap", gizmoMaterials.zTexture );
    gizmoMaterials.zAxis->SetVector("tint", Vec4( 0.2f, 0.2f, 0.2f, 1 ) );

    gizmoMaterials.xAxisHighlight = dll->MakeMaterial();
    gizmoMaterials.xAxisHighlight->SetShaders( "unlit", "unlit", "", "" );
    gizmoMaterials.xAxisHighlight->SetTexture("textureMap", gizmoMaterials.xTexture );
    gizmoMaterials.xAxisHighlight->SetVector("tint", Vec4( 1, 1, 1, 1) );

    gizmoMaterials.yAxisHighlight = dll->MakeMaterial();
    gizmoMaterials.yAxisHighlight->SetShaders( "unlit", "unlit", "", "" );
    gizmoMaterials.yAxisHighlight->SetTexture("textureMap", gizmoMaterials.yTexture );
    gizmoMaterials.yAxisHighlight->SetVector("tint", Vec4( 1, 1, 1, 1 ) );

    gizmoMaterials.zAxisHighlight = dll->MakeMaterial();
    gizmoMaterials.zAxisHighlight->SetShaders( "unlit", "unlit", "", "" );
    gizmoMaterials.zAxisHighlight->SetTexture("textureMap", gizmoMaterials.zTexture );
    gizmoMaterials.zAxisHighlight->SetVector("tint", Vec4( 1, 1, 1, 1 ) );

    cursorTranslate->SetMaterialForMesh( gizmoMaterials.xAxis, "xaxis" );
    cursorTranslate->SetMaterialForMesh( gizmoMaterials.yAxis, "yaxis" );
    cursorTranslate->SetMaterialForMesh( gizmoMaterials.zAxis, "zaxis" );
}

void MyGLWidget::SelectionPositionChanged( const Vec3& newPosition )
{
    cursorTranslate->SetPosition( newPosition );
}

void MyGLWidget::SetSelectedNodePosition( const Vec3& newPosition )
{
    editorState->selectedNode->SetPosition( newPosition );
}

bool MyGLWidget::HandleDrag( const QMouseEvent* mouseEvent, float deltaX, float deltaY )
{
    const float dragFactor = 25;

    if (dragAxis == GizmoAxis::X && mouseEvent->x() > desktop.geometry().width() - 4)
    {
        QCursor::setPos( 4, mouseEvent->y() );
        deltaX = 0;
        lastMousePosition[ 0 ] = 0;
        return true;
    }
    else if (dragAxis == GizmoAxis::Y && mouseEvent->y() > desktop.geometry().height() - 50)
    {
        QCursor::setPos( mouseEvent->y(), 4 );
        deltaY = 0;
        lastMousePosition[ 1 ] = 0;
        return true;
    }

    if (dragAxis == GizmoAxis::X && editorState->selectedNode != nullptr)
    {
        const float direction = editorCamera->ViewDirection().z > 0 ? 1 : -1;
        const Vec3 newPosition = editorState->selectedNode->GetPosition() + Vec3( -deltaX / dragFactor * direction, 0, 0 );
        SetSelectedNodePosition( newPosition );

        mainWindow->UpdateInspector();
        cursorTranslate->SetPosition( newPosition );
        lastMousePosition[ 0 ] = mouseEvent->x();
        lastMousePosition[ 1 ] = mouseEvent->y();
        return true;
    }
    else if (dragAxis == GizmoAxis::Y && editorState->selectedNode != nullptr)
    {
        const Vec3 newPosition = editorState->selectedNode->GetPosition() + Vec3( 0, deltaY / dragFactor, 0 );
        SetSelectedNodePosition( newPosition );
        mainWindow->UpdateInspector();
        cursorTranslate->SetPosition( newPosition );
        lastMousePosition[ 0 ] = mouseEvent->x();
        lastMousePosition[ 1 ] = mouseEvent->y();
        return true;
    }
    else if (dragAxis == GizmoAxis::Z && editorState->selectedNode != nullptr)
    {
        const float direction = editorCamera->ViewDirection().x > 0 ? -1 : 1;
        const Vec3 newPosition = editorState->selectedNode->GetPosition() + Vec3( 0, 0, -deltaX / dragFactor * direction );
        SetSelectedNodePosition( newPosition );
        mainWindow->UpdateInspector();
        cursorTranslate->SetPosition( newPosition );
        lastMousePosition[ 0 ] = mouseEvent->x();
        lastMousePosition[ 1 ] = mouseEvent->y();
        return true;
    }

    return false;
}

bool MyGLWidget::eventFilter(QObject */*obj*/, QEvent *event)
{
    if (event->type() == QEvent::MouseMove)
    {
        const QMouseEvent* mouseEvent = static_cast<QMouseEvent*>(event);

        float deltaX = (lastMousePosition[ 0 ] - mouseEvent->x()) * 0.1f;
        float deltaY = (lastMousePosition[ 1 ] - mouseEvent->y()) * 0.1f;

        if (grabMouse)
        {
            QPoint globalPos = mapToGlobal(QPoint(mouseEvent->x(), mouseEvent->y()));

            int x = mouseEvent->x();
            int y = mouseEvent->y();

            if (globalPos.x() < 5)
            {
                x = desktop.geometry().width() - 10;
                cursor().setPos( desktop.geometry().width() - 10, globalPos.y() );
            }
            else if (globalPos.x() > desktop.geometry().width() - 5)
            {
                x = 10;
                cursor().setPos( 10, globalPos.y() );
            }

            if (globalPos.y() < 5)
            {
                cursor().setPos( globalPos.x(), desktop.geometry().height() - 10 );
            }
            else if (globalPos.y() > desktop.geometry().height() - 10 )
            {
                cursor().setPos( globalPos.x(), 10 );
            }

            deltaX = deltaX > 5 ? 5 : deltaX;
            deltaX = deltaX < -5 ? -5 : deltaX;

            deltaY = deltaY > 5 ? 5 : deltaY;
            deltaY = deltaY < -5 ? -5 : deltaY;

            editorCamera->OffsetRotation( Vec3( 0.0f, 1.0f, 0.0f ), deltaX );
            editorCamera->OffsetRotation( Vec3( 1.0f, 0.0f, 0.0f ), deltaY );

            lastMousePosition[ 0 ] = x;
            lastMousePosition[ 1 ] = y;
            return true;
        }

        if (panMouse)
        {
            cameraMoveDir.x =  deltaX * 0.1f;
            cameraMoveDir.y = -deltaY * 0.1f;
            editorCamera->MoveRight(cameraMoveDir.x);
            editorCamera->MoveUp(cameraMoveDir.y);
            cameraMoveDir.x = cameraMoveDir.y = 0;
            lastMousePosition[ 0 ] = mouseEvent->x();
            lastMousePosition[ 1 ] = mouseEvent->y();
            return true;
        }

        if (HandleDrag( mouseEvent, deltaX, deltaY ))
        {
            return true;
        }

        HandleCursorHighlight();
    }

    return false;
}

void MyGLWidget::HandleCursorHighlight()
{
    QPoint point = mapFromGlobal( QCursor::pos() );
    auto colliders = scene->GetColliders( point.x(), point.y(), CollisionTest::AABB, 200 );

    const std::string xAxis = "xaxis";
    const std::string yAxis = "yaxis";
    const std::string zAxis = "zaxis";

    bool cursorInColliders = false;

    for (const auto& collider : colliders)
    {
        if (collider.first == cursorTranslate && !collider.second.empty())
        {
            cursorInColliders = true;
            break;
        }
    }

    if (cursorInColliders)
    {
        const std::string meshName = colliders.back().second.front();

        if (meshName == xAxis)
        {
            cursorTranslate->SetMaterialForMesh( gizmoMaterials.xAxisHighlight, xAxis );
            cursorTranslate->SetMaterialForMesh( gizmoMaterials.yAxis, yAxis );
            cursorTranslate->SetMaterialForMesh( gizmoMaterials.zAxis, zAxis );
        }
        else if (meshName == yAxis)
        {
            cursorTranslate->SetMaterialForMesh( gizmoMaterials.xAxis, xAxis );
            cursorTranslate->SetMaterialForMesh( gizmoMaterials.yAxisHighlight, yAxis );
            cursorTranslate->SetMaterialForMesh( gizmoMaterials.zAxis, zAxis );
        }
        else if (meshName == zAxis)
        {
            cursorTranslate->SetMaterialForMesh( gizmoMaterials.xAxis, xAxis );
            cursorTranslate->SetMaterialForMesh( gizmoMaterials.yAxis, yAxis );
            cursorTranslate->SetMaterialForMesh( gizmoMaterials.zAxisHighlight, zAxis );
        }
    }
    else
    {
        cursorTranslate->SetMaterialForMesh( gizmoMaterials.xAxis, xAxis );
        cursorTranslate->SetMaterialForMesh( gizmoMaterials.yAxis, yAxis );
        cursorTranslate->SetMaterialForMesh( gizmoMaterials.zAxis, zAxis );
    }
}

void MyGLWidget::DrawLightSprites()
{
    std::list< SceneGraphNode* > lights;
    GetSceneNodesByType( editorCamera->Root(), SceneGraphNode::Type::Light, lights );

    for (const auto& light : lights)
    {
        const float distance = (editorCamera->GetPosition() - light->GetPosition()).Length();
        const float lerpDistance = 10;
        float opacity = 1;

        if (distance < lerpDistance)
        {
            opacity = distance / lerpDistance;
        }

        const Vec3 screenPoint = editorCamera->ScreenPoint( light->GetPosition() );

        const Vec3 viewDir = editorCamera->ViewDirection();
        const Vec3 lightDir = (light->GetPosition() - editorCamera->GetPosition()).Normalized();
        const float viewDotLight = Vec3::Dot( viewDir, lightDir ) ;

        if (viewDotLight <= 0 &&
            screenPoint.x > -lightTex->Width() && screenPoint.y > -lightTex->Height() &&
            screenPoint.x < renderer->Width() && screenPoint.y < renderer->Height())
        {
            const Vec3 color = static_cast<Light*>( light )->GetColor();
            const float size = renderer->Height() / distance;

            renderer->Draw(lightTex, screenPoint.x, screenPoint.y, size, size, Vec4( color.x, color.y, color.z, opacity ) );
        }
    }
}

void MyGLWidget::DrawCameraSprites()
{
    std::list< SceneGraphNode* > cameras;
    GetSceneNodesByType( editorCamera->Root(), SceneGraphNode::Type::Camera, cameras );

    for (const auto& camera : cameras)
    {
        const float distance = (editorCamera->GetPosition() - camera->GetPosition()).Length();
        const float lerpDistance = 10;
        float opacity = 1;

        if (distance < lerpDistance)
        {
            opacity = distance / lerpDistance;
        }

        const Vec3 screenPoint = editorCamera->ScreenPoint( camera->GetPosition() );

        const Vec3 viewDir = editorCamera->ViewDirection();
        const Vec3 cameraDir = (camera->GetPosition() - editorCamera->GetPosition()).Normalized();
        const float viewDotLight = Vec3::Dot( viewDir, cameraDir ) ;

        if (camera != editorCamera && viewDotLight <= 0 &&
            screenPoint.x > -lightTex->Width() && screenPoint.y > -lightTex->Height() &&
            screenPoint.x < renderer->Width() && screenPoint.y < renderer->Height())
        {
            const float size = renderer->Height() / distance;

            renderer->Draw(cameraTex, screenPoint.x, screenPoint.y, size, size, Vec4( 1, 1, 1, opacity ) );
        }
    }
}

void MyGLWidget::mousePressEvent( QMouseEvent* event )
{
    setFocus();

    if (event->button() == Qt::RightButton && !grabMouse)
    {
        grabMouse = true;
        setCursor( Qt::BlankCursor );
        buttonDownMousePosition[ 0 ] = event->x();
        buttonDownMousePosition[ 1 ] = event->y();
        lastMousePosition[ 0 ] = event->x();
        lastMousePosition[ 1 ] = event->y();
        return;
    }

    if (event->button() == Qt::MiddleButton)
    {
        panMouse = true;
        lastMousePosition[ 0 ] = event->x();
        lastMousePosition[ 1 ] = event->y();
        cursor().setShape( Qt::ClosedHandCursor );
        return;
    }

    QPoint point = mapFromGlobal( QCursor::pos() );
    auto colliders = scene->GetColliders( point.x(), point.y(), CollisionTest::AABB, 200 );

    if (event->button() == Qt::LeftButton)
    {
        TryToSelectSprite( point.x(), point.y() );
    }

    for (auto collider : colliders)
    {
        if (dragAxis == GizmoAxis::None &&
            IsEditorWidget( collider.first ) && !collider.second.empty())
        {
            buttonDownMousePosition[ 0 ] = event->x();
            buttonDownMousePosition[ 1 ] = event->y();
            lastMousePosition[ 0 ] = event->x();
            lastMousePosition[ 1 ] = event->y();
            const std::string axis = collider.second.front();
            //std::cout << "Hit mesh: " << axis << std::endl;

            if (axis == "xaxis")
            {
                dragAxis = GizmoAxis::X;
            }
            else if (axis == "yaxis")
            {
                dragAxis = GizmoAxis::Y;
            }
            if (axis == "zaxis")
            {
                dragAxis = GizmoAxis::Z;
            }
        }
    }
}

void MyGLWidget::wheelEvent(QWheelEvent *event)
{
    const float dir = event->angleDelta().y() < 0 ? -1 : 1;
    const float speed = dir;
    editorCamera->MoveForward( speed );
}

void MyGLWidget::mouseReleaseEvent( QMouseEvent* /*event*/ )
{    
    if (dragAxis != GizmoAxis::None && editorState->selectedNode != nullptr)
    {
        dragAxis = GizmoAxis::None;
        return;
    }

    if (grabMouse)
    {
        grabMouse = false;
        unsetCursor();
        return;
    }

    if (panMouse)
    {
        cameraMoveDir.x = 0;
        cameraMoveDir.y = 0;
        panMouse = false;
        cursor().setShape( Qt::ArrowCursor );

        return;
    }

    QPoint point = mapFromGlobal( QCursor::pos() );
    auto colliders = scene->GetColliders( point.x(), point.y(), CollisionTest::AABB, 200 );

    if (!colliders.empty() && !IsEditorWidget( colliders.back().first ))
    {
        //std::cout << "MyGLWidget: " << colliders.back().first->GetMeshNames()[ 0 ] << std::endl;
        mainWindow->SetSelectedNode( colliders.back().first );
        dragAxis = GizmoAxis::None;
    }
    else if (!colliders.empty() && IsEditorWidget( colliders.back().first ) && !colliders.back().second.empty())
    {
        //std::cout << "hit mesh count: " << colliders.front().second.size() << std::endl;
        std::cout << "Hit mesh: " << colliders.back().second.front() << std::endl;
    }
    else
    {
        mainWindow->SetSelectedNode( nullptr );
        dragAxis = GizmoAxis::None;
    }

    unsetCursor();
}

void MyGLWidget::TryToSelectSprite( int x, int y )
{
    std::list< SceneGraphNode* > lights;
    GetSceneNodesByType( editorCamera->Root(), SceneGraphNode::Type::Light, lights );

    for (const auto& light : lights)
    {
        const float distance = (editorCamera->GetPosition() - light->GetPosition()).Length();
        const float lerpDistance = 10;
        float opacity = 1;

        if (distance < lerpDistance)
        {
            opacity = distance / lerpDistance;
        }

        const Vec3 screenPoint = editorCamera->ScreenPoint( light->GetPosition() );
    }
}

Model* MyGLWidget::AddModel( const char* path )
{
    Error::Enum error = Error::NO_ERRORS;
    Model* mo = dll->MakeModel( path, error );
    mo->SetPosition( editorCamera->GetPosition() - editorCamera->ViewDirection() * 2);
    mo->SetMaterialForMeshSubstring( unlit, "" );
    mo->SetName( "model" );
    scene->Add( mo );
    mainWindow->SetSelectedNode( mo );
    return mo;
}

Light* MyGLWidget::AddLight( Light::Type type )
{
    Light* light = nullptr;
    light = dll->MakeLight( type );

    if (type == Light::Type::Point)
    {
        light->SetName( "Point Light" );
    }
    else if (type == Light::Type::Directional)
    {
        light->SetName( "Directional Light" );
    }
    else if (type == Light::Type::Spot)
    {
        light->SetName( "Spot Light" );
    }

    scene->Add( light );
    return light;
}

void MyGLWidget::AddEmpty()
{
    SceneGraphNode* node = new SceneGraphNode();
    node->SetName( "Empty Node" );
    node->SetPosition( editorCamera->GetPosition() - editorCamera->ViewDirection() * 2);
    mainWindow->SetSelectedNode( node );
    editorCamera->Root()->AddChild( node );
}

void MyGLWidget::AddCamera()
{
    Camera* camera = dll->MakeCamera();
    camera->SetProjection( 55, 16.0f / 9.0f, 1, 200 );
    camera->SetName( "Camera" );
    camera->SetRenderTarget( cameraRenderTexture );
    scene->Add( camera );
}

void MyGLWidget::PreSave()
{
    scene->Remove( cursorTranslate );
    scene->Remove( editorCamera );
}

void MyGLWidget::PostSave()
{
    scene->Add( cursorTranslate );
    scene->Add( editorCamera );
}

void MyGLWidget::RemoveNode( SceneGraphNode* node )
{
    scene->Remove( node );
    mainWindow->SetSelectedNode( nullptr );
    SelectionChanged();
}

void MyGLWidget::CenterSelected()
{
    assert( editorState->selectedNode );

    Vec3 targetPosition = editorState->selectedNode->GetPosition();
    //Vec3 startPosition = editorCamera->GetPosition();

    float zDistance = 1;

    if (editorState->selectedNode->GetType() == SceneGraphNode::Type::Model)
    {
        const Model* model = static_cast< Model* >( editorState->selectedNode );
        Vec3 aabbMin, aabbMax;
        model->GetAABB( aabbMin, aabbMax );
        zDistance = (aabbMax - aabbMin).z + 1;
    }

    editorCamera->LookAt( targetPosition - Vec3( 0, 0, zDistance ), targetPosition, Vec3( 0, 1, 0 ) );
}

void MyGLWidget::DeinitRenderer()
{
    delete cursorTranslate;
    delete sphere;
    delete editorCamera;
    delete unlit;
    delete checker;
    delete scene;
    renderer->Deinit();
}

void MyGLWidget::SelectionChanged()
{
    cursorTranslate->SetLayerMask( editorState->selectedNode == nullptr ? 0x0 : 0xFFFFFFFF );
}
